@@ -1,30 +1,33 @@
 # -*- coding: utf-8 -*-
-from __future__ import unicode_literals, print_function
-import six
-from six.moves.urllib.parse import urlparse, urlunparse
+from __future__ import print_function, unicode_literals
 
 import copy
 import locale
 import logging
-import functools
 import os
 import re
 import sys
 
 import pytz
 
+import six
+from six.moves.urllib.parse import urljoin, urlparse, urlunparse
+
 from pelican import signals
 from pelican.settings import DEFAULT_CONFIG
-from pelican.utils import (slugify, truncate_html_words, memoized, strftime,
-                           python_2_unicode_compatible, deprecated_attribute,
-                           path_to_url, posixize_path, set_date_tzinfo, SafeDatetime)
+from pelican.utils import (SafeDatetime, deprecated_attribute, memoized,
+                           path_to_url, posixize_path,
+                           python_2_unicode_compatible, sanitised_join,
+                           set_date_tzinfo, slugify, strftime,
+                           truncate_html_words)
 
 # Import these so that they're avalaible when you import from pelican.contents.
-from pelican.urlwrappers import (URLWrapper, Author, Category, Tag)  # NOQA
+from pelican.urlwrappers import (Author, Category, Tag, URLWrapper)  # NOQA
 
 logger = logging.getLogger(__name__)
 
 
+@python_2_unicode_compatible
 class Content(object):
     """Represents a content.
 
@@ -65,7 +68,7 @@ class Content(object):
         # also keep track of the metadata attributes available
         self.metadata = local_metadata
 
-        #default template if it's not defined in page
+        # default template if it's not defined in page
         self.template = self._get_template()
 
         # First, read the authors from "authors", if not, fallback to "author"
@@ -93,13 +96,18 @@ class Content(object):
         # create the slug if not existing, generate slug according to
         # setting of SLUG_ATTRIBUTE
         if not hasattr(self, 'slug'):
-            if settings['SLUGIFY_SOURCE'] == 'title' and hasattr(self, 'title'):
-                self.slug = slugify(self.title,
-                                settings.get('SLUG_SUBSTITUTIONS', ()))
-            elif settings['SLUGIFY_SOURCE'] == 'basename' and source_path != None:
-                basename = os.path.basename(os.path.splitext(source_path)[0])
-                self.slug = slugify(basename,
-                                settings.get('SLUG_SUBSTITUTIONS', ()))
+            if (settings['SLUGIFY_SOURCE'] == 'title' and
+                    hasattr(self, 'title')):
+                self.slug = slugify(
+                    self.title,
+                    regex_subs=settings.get('SLUG_REGEX_SUBSTITUTIONS', []))
+            elif (settings['SLUGIFY_SOURCE'] == 'basename' and
+                    source_path is not None):
+                basename = os.path.basename(
+                    os.path.splitext(source_path)[0])
+                self.slug = slugify(
+                    basename,
+                    regex_subs=settings.get('SLUG_REGEX_SUBSTITUTIONS', []))
 
         self.source_path = source_path
 
@@ -132,14 +140,7 @@ class Content(object):
 
         # manage status
         if not hasattr(self, 'status'):
-            self.status = settings['DEFAULT_STATUS']
-            if not settings['WITH_FUTURE_DATES'] and hasattr(self, 'date'):
-                if self.date.tzinfo is None:
-                    now = SafeDatetime.now()
-                else:
-                    now = SafeDatetime.utcnow().replace(tzinfo=pytz.utc)
-                if self.date > now:
-                    self.status = 'draft'
+            self.status = getattr(self, 'default_status', None)
 
         # store the summary metadata if it is set
         if 'summary' in metadata:
@@ -148,18 +149,58 @@ class Content(object):
         signals.content_object_init.send(self)
 
     def __str__(self):
-        if self.source_path is None:
-            return repr(self)
-        elif six.PY3:
-            return self.source_path or repr(self)
-        else:
-            return str(self.source_path.encode('utf-8', 'replace'))
+        return self.source_path or repr(self)
 
-    def check_properties(self):
+    def _has_valid_mandatory_properties(self):
         """Test mandatory properties are set."""
         for prop in self.mandatory_properties:
             if not hasattr(self, prop):
-                raise NameError(prop)
+                logger.error(
+                    "Skipping %s: could not find information about '%s'",
+                    self, prop)
+                return False
+        return True
+
+    def _has_valid_save_as(self):
+        """Return true if save_as doesn't write outside output path, false
+        otherwise."""
+        try:
+            output_path = self.settings["OUTPUT_PATH"]
+        except KeyError:
+            # we cannot check
+            return True
+
+        try:
+            sanitised_join(output_path, self.save_as)
+        except RuntimeError:  # outside output_dir
+            logger.error(
+                "Skipping %s: file %r would be written outside output path",
+                self,
+                self.save_as,
+            )
+            return False
+
+        return True
+
+    def _has_valid_status(self):
+        if hasattr(self, 'allowed_statuses'):
+            if self.status not in self.allowed_statuses:
+                logger.error(
+                    "Unknown status '%s' for file %s, skipping it.",
+                    self.status,
+                    self
+                )
+                return False
+
+        # if undefined we allow all
+        return True
+
+    def is_valid(self):
+        """Validate Content"""
+        # Use all() to not short circuit and get results of all validations
+        return all([self._has_valid_mandatory_properties(),
+                    self._has_valid_save_as(),
+                    self._has_valid_status()])
 
     @property
     def url_format(self):
@@ -172,12 +213,15 @@ class Content(object):
             'lang': getattr(self, 'lang', 'en'),
             'date': getattr(self, 'date', SafeDatetime.now()),
             'author': self.author.slug if hasattr(self, 'author') else '',
+            'tag': self.tag.slug if hasattr(self, 'tag') else '',
             'category': self.category.slug if hasattr(self, 'category') else ''
         })
         return metadata
 
-    def _expand_settings(self, key):
-        fq_key = ('%s_%s' % (self.__class__.__name__, key)).upper()
+    def _expand_settings(self, key, klass=None):
+        if not klass:
+            klass = self.__class__.__name__
+        fq_key = ('%s_%s' % (klass, key)).upper()
         return self.settings[fq_key].format(**self.url_format)
 
     def get_url_setting(self, key):
@@ -186,6 +230,109 @@ class Content(object):
         key = key if self.in_default_lang else 'lang_%s' % key
         return self._expand_settings(key)
 
+    def _link_replacer(self, siteurl, m):
+        what = m.group('what')
+        value = urlparse(m.group('value'))
+        path = value.path
+        origin = m.group('path')
+
+        # urllib.parse.urljoin() produces `a.html` for urljoin("..", "a.html")
+        # so if RELATIVE_URLS are enabled, we fall back to os.path.join() to
+        # properly get `../a.html`. However, os.path.join() produces
+        # `baz/http://foo/bar.html` for join("baz", "http://foo/bar.html")
+        # instead of correct "http://foo/bar.html", so one has to pick a side
+        # as there is no silver bullet.
+        if self.settings['RELATIVE_URLS']:
+            joiner = os.path.join
+        else:
+            joiner = urljoin
+
+            # However, it's not *that* simple: urljoin("blog", "index.html")
+            # produces just `index.html` instead of `blog/index.html` (unlike
+            # os.path.join()), so in order to get a correct answer one needs to
+            # append a trailing slash to siteurl in that case. This also makes
+            # the new behavior fully compatible with Pelican 3.7.1.
+            if not siteurl.endswith('/'):
+                siteurl += '/'
+
+        # XXX Put this in a different location.
+        if what in {'filename', 'static', 'attach'}:
+            if path.startswith('/'):
+                path = path[1:]
+            else:
+                # relative to the source path of this content
+                path = self.get_relative_source_path(
+                    os.path.join(self.relative_dir, path)
+                )
+
+            key = 'static_content' if what in ('static', 'attach')\
+                else 'generated_content'
+
+            def _get_linked_content(key, path):
+                try:
+                    return self._context[key][path]
+                except KeyError:
+                    try:
+                        # Markdown escapes spaces, try unescaping
+                        return self._context[key][path.replace('%20', ' ')]
+                    except KeyError:
+                        if what == 'filename' and key == 'generated_content':
+                            key = 'static_content'
+                            linked_content = _get_linked_content(key, path)
+                            if linked_content:
+                                logger.warning(
+                                    '{filename} used for linking to static'
+                                    'content %s in %s. Use {static} instead',
+                                    path,
+                                    self.get_relative_source_path())
+                                return linked_content
+                        return None
+
+            linked_content = _get_linked_content(key, path)
+            if linked_content:
+                if what == 'attach':
+                    linked_content.attach_to(self)
+                origin = joiner(siteurl, linked_content.url)
+                origin = origin.replace('\\', '/')  # for Windows paths.
+            else:
+                logger.warning(
+                    "Unable to find '%s', skipping url replacement.",
+                    value.geturl(), extra={
+                        'limit_msg': ("Other resources were not found "
+                                      "and their urls not replaced")})
+        elif what == 'category':
+            origin = joiner(siteurl, Category(path, self.settings).url)
+        elif what == 'tag':
+            origin = joiner(siteurl, Tag(path, self.settings).url)
+        elif what == 'index':
+            origin = joiner(siteurl, self.settings['INDEX_SAVE_AS'])
+        elif what == 'author':
+            origin = joiner(siteurl, Author(path, self.settings).url)
+        else:
+            logger.warning(
+                "Replacement Indicator '%s' not recognized, "
+                "skipping replacement",
+                what)
+
+        # keep all other parts, such as query, fragment, etc.
+        parts = list(value)
+        parts[2] = origin
+        origin = urlunparse(parts)
+
+        return ''.join((m.group('markup'), m.group('quote'), origin,
+                        m.group('quote')))
+
+    def _get_intrasite_link_regex(self):
+        intrasite_link_regex = self.settings['INTRASITE_LINK_REGEX']
+        regex = r"""
+            (?P<markup><[^\>]+  # match tag with all url-value attributes
+                (?:href|src|poster|data|cite|formaction|action)\s*=\s*)
+
+            (?P<quote>["\'])      # require value to be quoted
+            (?P<path>{0}(?P<value>.*?))  # the url value
+            \2""".format(intrasite_link_regex)
+        return re.compile(regex, re.X)
+
     def _update_content(self, content, siteurl):
         """Update the content attribute.
 
@@ -199,69 +346,31 @@ class Content(object):
         if not content:
             return content
 
-        instrasite_link_regex = self.settings['INTRASITE_LINK_REGEX']
-        regex = r"""
-            (?P<markup><\s*[^\>]*  # match tag with all url-value attributes
-                (?:href|src|poster|data|cite|formaction|action)\s*=)
+        hrefs = self._get_intrasite_link_regex()
+        return hrefs.sub(lambda m: self._link_replacer(siteurl, m), content)
 
-            (?P<quote>["\'])      # require value to be quoted
-            (?P<path>{0}(?P<value>.*?))  # the url value
-            \2""".format(instrasite_link_regex)
-        hrefs = re.compile(regex, re.X)
-
-        def replacer(m):
+    def get_static_links(self):
+        static_links = set()
+        hrefs = self._get_intrasite_link_regex()
+        for m in hrefs.finditer(self._content):
             what = m.group('what')
             value = urlparse(m.group('value'))
             path = value.path
-            origin = m.group('path')
+            if what not in {'static', 'attach'}:
+                continue
+            if path.startswith('/'):
+                path = path[1:]
+            else:
+                # relative to the source path of this content
+                path = self.get_relative_source_path(
+                    os.path.join(self.relative_dir, path)
+                )
+            path = path.replace('%20', ' ')
+            static_links.add(path)
+        return static_links
 
-            # XXX Put this in a different location.
-            if what in {'filename', 'attach'}:
-                if path.startswith('/'):
-                    path = path[1:]
-                else:
-                    # relative to the source path of this content
-                    path = self.get_relative_source_path(
-                        os.path.join(self.relative_dir, path)
-                    )
-
-                if path not in self._context['filenames']:
-                    unquoted_path = path.replace('%20', ' ')
-
-                    if unquoted_path in self._context['filenames']:
-                        path = unquoted_path
-
-                linked_content = self._context['filenames'].get(path)
-                if linked_content:
-                    if what == 'attach':
-                        if isinstance(linked_content, Static):
-                            linked_content.attach_to(self)
-                        else:
-                            logger.warning("%s used {attach} link syntax on a "
-                                "non-static file. Use {filename} instead.",
-                                self.get_relative_source_path())
-                    origin = '/'.join((siteurl, linked_content.url))
-                    origin = origin.replace('\\', '/')  # for Windows paths.
-                else:
-                    logger.warning(
-                        "Unable to find `%s`, skipping url replacement.",
-                        value.geturl(), extra = {
-                            'limit_msg': ("Other resources were not found "
-                                          "and their urls not replaced")})
-            elif what == 'category':
-                origin = '/'.join((siteurl, Category(path, self.settings).url))
-            elif what == 'tag':
-                origin = '/'.join((siteurl, Tag(path, self.settings).url))
-
-            # keep all other parts, such as query, fragment, etc.
-            parts = list(value)
-            parts[2] = origin
-            origin = urlunparse(parts)
-
-            return ''.join((m.group('markup'), m.group('quote'), origin,
-                            m.group('quote')))
-
-        return hrefs.sub(replacer, content)
+    def get_siteurl(self):
+        return self._context.get('localsiteurl', '')
 
     @memoized
     def get_content(self, siteurl):
@@ -271,22 +380,19 @@ class Content(object):
             content = self._content
         return self._update_content(content, siteurl)
 
-    def get_siteurl(self):
-        return self._context.get('localsiteurl', '')
-
     @property
     def content(self):
         return self.get_content(self.get_siteurl())
 
-    def _get_summary(self):
+    @memoized
+    def get_summary(self, siteurl):
         """Returns the summary of an article.
 
         This is based on the summary metadata if set, otherwise truncate the
         content.
         """
-        if hasattr(self, '_summary'):
-            return self._update_content(self._summary,
-                                        self.get_siteurl())
+        if 'summary' in self.metadata:
+            return self.metadata['summary']
 
         if self.settings['SUMMARY_MAX_LENGTH'] is None:
             return self.content
@@ -294,20 +400,31 @@ class Content(object):
         return truncate_html_words(self.content,
                                    self.settings['SUMMARY_MAX_LENGTH'])
 
-    @memoized
-    def get_summary(self, siteurl):
-        """uses siteurl to be memoizable"""
-        return self._get_summary()
-
     @property
     def summary(self):
         return self.get_summary(self.get_siteurl())
 
+    def _get_summary(self):
+        """deprecated function to access summary"""
+
+        logger.warning('_get_summary() has been deprecated since 3.6.4. '
+                       'Use the summary decorator instead')
+        return self.summary
+
     @summary.setter
     def summary(self, value):
         """Dummy function"""
         pass
 
+    @property
+    def status(self):
+        return self._status
+
+    @status.setter
+    def status(self, value):
+        # TODO maybe typecheck
+        self._status = value.lower()
+
     @property
     def url(self):
         return self.get_url_setting('url')
@@ -336,7 +453,9 @@ class Content(object):
 
         return posixize_path(
             os.path.relpath(
-                os.path.abspath(os.path.join(self.settings['PATH'], source_path)),
+                os.path.abspath(os.path.join(
+                    self.settings['PATH'],
+                    source_path)),
                 os.path.abspath(self.settings['PATH'])
             ))
 
@@ -348,24 +467,63 @@ class Content(object):
                     os.path.abspath(self.source_path),
                     os.path.abspath(self.settings['PATH']))))
 
+    def refresh_metadata_intersite_links(self):
+        for key in self.settings['FORMATTED_FIELDS']:
+            if key in self.metadata and key != 'summary':
+                value = self._update_content(
+                    self.metadata[key],
+                    self.get_siteurl()
+                )
+                self.metadata[key] = value
+                setattr(self, key.lower(), value)
+
+        # _summary is an internal variable that some plugins may be writing to,
+        # so ensure changes to it are picked up
+        if ('summary' in self.settings['FORMATTED_FIELDS'] and
+                'summary' in self.metadata):
+            self._summary = self._update_content(
+                self._summary,
+                self.get_siteurl()
+            )
+            self.metadata['summary'] = self._summary
+
 
 class Page(Content):
     mandatory_properties = ('title',)
+    allowed_statuses = ('published', 'hidden', 'draft')
+    default_status = 'published'
     default_template = 'page'
 
+    def _expand_settings(self, key):
+        klass = 'draft_page' if self.status == 'draft' else None
+        return super(Page, self)._expand_settings(key, klass)
 
-class Article(Page):
+
+class Article(Content):
     mandatory_properties = ('title', 'date', 'category')
+    allowed_statuses = ('published', 'draft')
+    default_status = 'published'
     default_template = 'article'
 
+    def __init__(self, *args, **kwargs):
+        super(Article, self).__init__(*args, **kwargs)
 
-class Draft(Page):
-    mandatory_properties = ('title', 'category')
-    default_template = 'article'
+        # handle WITH_FUTURE_DATES (designate article to draft based on date)
+        if not self.settings['WITH_FUTURE_DATES'] and hasattr(self, 'date'):
+            if self.date.tzinfo is None:
+                now = SafeDatetime.now()
+            else:
+                now = SafeDatetime.utcnow().replace(tzinfo=pytz.utc)
+            if self.date > now:
+                self.status = 'draft'
 
+        # if we are a draft and there is no date provided, set max datetime
+        if not hasattr(self, 'date') and self.status == 'draft':
+            self.date = SafeDatetime.max
 
-class Quote(Page):
-    base_properties = ('author', 'date')
+    def _expand_settings(self, key):
+        klass = 'draft' if self.status == 'draft' else 'article'
+        return super(Article, self)._expand_settings(key, klass)
 
 
 @python_2_unicode_compatible
@@ -401,9 +559,12 @@ class Static(Page):
     def attach_to(self, content):
         """Override our output directory with that of the given content object.
         """
-        # Determine our file's new output path relative to the linking document.
-        # If it currently lives beneath the linking document's source directory,
-        # preserve that relationship on output. Otherwise, make it a sibling.
+
+        # Determine our file's new output path relative to the linking
+        # document. If it currently lives beneath the linking
+        # document's source directory, preserve that relationship on output.
+        # Otherwise, make it a sibling.
+
         linking_source_dir = os.path.dirname(content.source_path)
         tail_path = os.path.relpath(self.source_path, linking_source_dir)
         if tail_path.startswith(os.pardir + os.sep):
@@ -419,11 +580,14 @@ class Static(Page):
         # 'some/content' with a file named 'index.html'.) Rather than trying
         # to figure it out by comparing the linking document's url and save_as
         # path, we simply build our new url from our new save_as path.
+
         new_url = path_to_url(new_save_as)
 
         def _log_reason(reason):
-            logger.warning("The {attach} link in %s cannot relocate %s "
-                "because %s. Falling back to {filename} link behavior instead.",
+            logger.warning(
+                "The {attach} link in %s cannot relocate "
+                "%s because %s. Falling back to "
+                "{filename} link behavior instead.",
                 content.get_relative_source_path(),
                 self.get_relative_source_path(), reason,
                 extra={'limit_msg': "More {attach} warnings silenced."})
@@ -444,12 +608,3 @@ class Static(Page):
 
         self.override_save_as = new_save_as
         self.override_url = new_url
-
-
-def is_valid_content(content, f):
-    try:
-        content.check_properties()
-        return True
-    except NameError as e:
-        logger.error("Skipping %s: could not find information about '%s'", f, e)
-        return False
